// GDIPlusHelper.cpp: implementation of the CGDIPlusHelper class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "AnimImage.h"
#include <process.h>

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CAnimImage::CAnimImage(LPCTSTR sResourceType, LPCTSTR sResource)
{
	Initialize();

	if (Load(sResourceType, sResource))
	{
	   nativeImage = NULL;	
       if (m_pStream) //lint !e1542
       {
           lastResult = DllExports::GdipLoadImageFromStreamICM(m_pStream, &nativeImage);
           TestForAnimatedGIF();
       }
    }
}

CAnimImage::CAnimImage(const WCHAR* filename, BOOL useEmbeddedColorManagement) : Image(filename, useEmbeddedColorManagement)
{
	Initialize();

	m_bIsInitialized = true;

	TestForAnimatedGIF();
}

CAnimImage::~CAnimImage()
{
	Destroy();
}

void CAnimImage::Initialize()
{
	m_pStream = NULL;
	m_nFramePosition = 0;
	m_nFrameCount = 0;
	m_pStream = NULL;
	lastResult = InvalidParameter;
	m_hThread = NULL;
	m_bIsInitialized = FALSE;
	m_pPropertyItem = NULL;

	m_hInst = AfxGetResourceHandle();

	m_bPause = FALSE;

	m_hExitEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

	m_hPause = CreateEvent(NULL, TRUE, TRUE, NULL);
	m_hWnd = NULL;
}

BOOL CAnimImage::InitAnimation(HWND hWnd, CPoint pt)
{
	m_hWnd = hWnd;
	m_pt = pt;

	if (!m_bIsInitialized)
	{
		return FALSE;
	}

	if (IsAnimatedGIF())
	{
		if (m_hThread == NULL)
		{		
			unsigned int nTID = 0;

			m_hThread = (HANDLE)_beginthreadex(NULL, 0, _ThreadAnimationProc, this, CREATE_SUSPENDED, &nTID);
			
			if (!m_hThread)
			{
				return TRUE;
			} 
			else 
				ResumeThread(m_hThread);
		}
	} 

	return FALSE;	

}

BOOL CAnimImage::LoadFromBuffer(BYTE* pBuff, int nSize)
{
	BOOL bResult = FALSE;

	HGLOBAL hGlobal = GlobalAlloc(GMEM_MOVEABLE, nSize);
	if (hGlobal)
	{
		void* pData = GlobalLock(hGlobal);
		if (pData)
			memcpy(pData, pBuff, nSize);
		
		GlobalUnlock(hGlobal);

		if (CreateStreamOnHGlobal(hGlobal, TRUE, &m_pStream) == S_OK)
			bResult = TRUE;
	}

	return bResult;
}

BOOL CAnimImage::GetResource(LPCTSTR lpName, LPCTSTR lpType, void* pResource, int& nBufSize)
{ 
	HRSRC		hResInfo;
	HANDLE		hRes;
	LPSTR		lpRes	= NULL; 
	//int			nLen	= 0;
	bool		bResult	= FALSE;

	hResInfo = FindResource(m_hInst , lpName, lpType);
	if (hResInfo == NULL) 
	{
		DWORD dwErr = GetLastError();
		return FALSE;
	}

	hRes = LoadResource(m_hInst , hResInfo);

	if (hRes == NULL) 
		return FALSE;

	lpRes = (char*)LockResource(hRes);
	if (lpRes != NULL)
	{ 
		if (pResource == NULL)
		{
			nBufSize = SizeofResource(m_hInst , hResInfo);
			bResult = TRUE;
		}
		else
		{
			if (nBufSize >= (int)SizeofResource(m_hInst , hResInfo))
			{
				memcpy(pResource, lpRes, nBufSize);
				bResult = TRUE;
			}
		} 

		UnlockResource(hRes);  
	}

	FreeResource(hRes);

	return bResult;
}

BOOL CAnimImage::Load(CString sResourceType, CString sResource)
{
	BOOL bResult = FALSE;

	BYTE*	pBuff = NULL;
	int		nSize = 0;
	if (GetResource(sResource.GetBuffer(0), sResourceType.GetBuffer(0), pBuff, nSize))
	{
		if (nSize > 0)
		{
			pBuff = new BYTE[nSize];
			if (NULL != pBuff)
			{
				if (GetResource(sResource, sResourceType.GetBuffer(0), pBuff, nSize))
				{
					if (LoadFromBuffer(pBuff, nSize))
					{
						bResult = TRUE;
					}
				}

				delete [] pBuff;
			}
		}
	}

	m_bIsInitialized = bResult;

	return bResult;
}

CSize CAnimImage::GetSize()
{
	return CSize(GetWidth(), GetHeight());
}

BOOL CAnimImage::TestForAnimatedGIF()
{
   UINT count = 0;
   count = GetFrameDimensionsCount();
   GUID* pDimensionIDs = new GUID[count];
   if (NULL == pDimensionIDs)
   {
	   return FALSE;
   }

   GetFrameDimensionsList(pDimensionIDs, count);

   m_nFrameCount = GetFrameCount(&pDimensionIDs[0]);
   int nSize = GetPropertyItemSize(PropertyTagFrameDelay);

   m_pPropertyItem = (PropertyItem*)malloc(nSize);
   if (NULL == m_pPropertyItem)
   {
	   delete [] pDimensionIDs;
	   return FALSE;
   }

   GetPropertyItem(PropertyTagFrameDelay, nSize, m_pPropertyItem);
   
   delete [] pDimensionIDs;

   return m_nFrameCount > 1;
}

UINT WINAPI CAnimImage::_ThreadAnimationProc(LPVOID pParam)
{
	ASSERT(pParam);
	CAnimImage* pImage = reinterpret_cast<CAnimImage*>(pParam);
	pImage->ThreadAnimation();

	return 0;
}

void CAnimImage::ThreadAnimation()
{
	m_nFramePosition = 0;

	BOOL bExit = FALSE;
	while (bExit == FALSE)
	{
		bExit = DrawFrameGIF();
	}
}

BOOL CAnimImage::DrawFrameGIF()
{
	::WaitForSingleObject(m_hPause, INFINITE);

	GUID pageGuid = FrameDimensionTime;

	long hmWidth = GetWidth();
	long hmHeight = GetHeight();

	HDC hDC = GetDC(m_hWnd);
	if (hDC)
	{
		Graphics graphics(hDC);
		graphics.DrawImage(this, m_pt.x, m_pt.y, hmWidth, hmHeight);
		ReleaseDC(m_hWnd, hDC);
	}

	SelectActiveFrame(&pageGuid, m_nFramePosition++);		
	
	if (m_nFramePosition == m_nFrameCount)
		m_nFramePosition = 0;

	long lPause = 0;
	if (NULL != m_pPropertyItem)
	{
		lPause = ((long*) m_pPropertyItem->value)[m_nFramePosition] * 10;
	}

	DWORD dwErr = WaitForSingleObject(m_hExitEvent, lPause);

	return (dwErr == WAIT_OBJECT_0);
}

void CAnimImage::SetPause(BOOL bPause)
{
	if (!IsAnimatedGIF())
		return;

	if (bPause && !m_bPause)
	{
		::ResetEvent(m_hPause);
	}
	else
	{
		if (m_bPause && !bPause)
		{
			::SetEvent(m_hPause);
		}
	}

	m_bPause = bPause;
}

void CAnimImage::Destroy()
{	
	if (m_hThread)
	{
		SetPause(false);

		SetEvent(m_hExitEvent);
		WaitForSingleObject(m_hThread, INFINITE);
	}

	CloseHandle(m_hThread);
	CloseHandle(m_hExitEvent);
	CloseHandle(m_hPause);

	if (NULL != m_pPropertyItem)
    {
        free(m_pPropertyItem);
    	m_pPropertyItem = NULL;
    }
	m_hThread = NULL;
	m_hExitEvent = NULL;
	m_hPause = NULL;
    m_hInst = NULL;
    m_hWnd = NULL;

    if (m_pStream){
		m_pStream->Release();
        m_pStream = NULL;
    }
}
